#include "UDPSimple.hpp"
#include <boost/algorithm/string.hpp>
#include <fstream>


const std::string currentDateTime();
void copy_delim_string_as_cf32(std::vector<std::complex<float> > &dest_signal, std::string source_delim_str);
void read_binary_file_as_cf32(std::vector<std::complex<float> > &dest_vector, std::string source_file_str);
void write_cf32_to_binary_file(std::string dest_file_str, void *ptr_cf32, unsigned int n_cf32_values);
void plot_cf32_buffer(std::string udp_dst_addr, std::string port, void *ptr_cf32, unsigned int n_cf32_values);

// Get current date/time, format is YYYY-MM-DD.HH-mm-ss
const std::string currentDateTime() {
  time_t     now = time(0);
  struct tm  tstruct;
  char       buf[80];
  tstruct = *localtime(&now);
 
 // Visit http://en.cppreference.com/w/cpp/chrono/c/strftime
  // for more information about date/time format
  strftime(buf, sizeof(buf), "%Y-%m-%d.%H-%M-%S", &tstruct);

  return buf;
}

void copy_delim_string_as_cf32(std::vector<std::complex<float> > &dest_signal, std::string source_delim_str)
{
  std::vector<std::string> tokens;
  boost::split(tokens, source_delim_str, boost::is_any_of("| "), boost::token_compress_on);
  for (unsigned int i = 0; i < tokens.size() ; i++) {
    std::complex<float> cf = boost::lexical_cast<std::complex<float> >( tokens.at(i) );
    dest_signal.push_back(cf);
    //std::cerr << tokens.at(i) << " " << cf << std::endl;
  }
  assert(tokens.size() == dest_signal.size() );
}

void read_binary_file_as_cf32(std::vector<std::complex<float> > &dest_vector, std::string source_file_str)
{
  std::ifstream is (source_file_str.c_str(), std::ifstream::binary);
  is.seekg (0, is.end);
  unsigned int length = is.tellg();
  is.seekg (0, is.beg);
  unsigned int nSamples = length / sizeof(std::complex<float>);
  dest_vector.resize( nSamples );
  is.read ( (char*)dest_vector.data(), length);
}

void write_cf32_to_binary_file(std::string dest_file_str, void *ptr_cf32, unsigned int n_cf32_values)
{
  std::ofstream outs (dest_file_str.c_str(), /*std::ofstream::out | */std::ofstream::binary);
  outs.write((char*) ptr_cf32, n_cf32_values * sizeof(std::complex<float>) );
  outs.close();
}

void plot_cf32_buffer(std::string udp_dst_addr, std::string port, void *ptr_cf32, unsigned int n_cf32_values)
{
  static UDPSimple udp_xport(udp_dst_addr, port);

  // send number of channels (or streams)
  unsigned int Channels = 1;
  udp_xport.send( (void*)&Channels, sizeof(Channels) );

  // send number of cf32 samples to receive
  udp_xport.send( (void*)&n_cf32_values, sizeof(n_cf32_values) );

  unsigned int _spb = 256;
  if (n_cf32_values < 256) _spb = n_cf32_values;

  unsigned int num_samps_per_datagram = _spb; //num_samps_per_datagram = (uhd::transport::udp_simple::mtu) / sizeof(std::complex<float>);

  // send number of datagrams
  unsigned int num_udp_datagrams = floor((float)n_cf32_values / (float)(num_samps_per_datagram));
  udp_xport.send((void*)&num_udp_datagrams, sizeof(num_udp_datagrams));

  std::complex<float> *ptr_to_buffer = (std::complex<float> *)ptr_cf32;
  for (unsigned int i = 0; i < num_udp_datagrams; ++i) {
    udp_xport.send( (void*)ptr_to_buffer, num_samps_per_datagram * sizeof(std::complex<float>) );
    ptr_to_buffer += num_samps_per_datagram;
    //boost::this_thread::sleep(boost::posix_time::milliseconds(10));
  }
}

