#include <iostream>
#include <complex>

//#include <fstream>
//#include <sstream>

//#include <fftw3.h>

#include <boost/program_options.hpp>
#include <boost/format.hpp>
#include <boost/thread.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/asio.hpp>

#include "CDeviceStorage.hpp"
#include "CRadio.hpp"

#include "log4cxx/logger.h"
#include "log4cxx/propertyconfigurator.h"
#include "log4cxx/helpers/exception.h"

#include <signal.h>


#define _USING_WEBSOCK_
#ifdef _USING_WEBSOCK_
#include <websocketpp/config/asio_no_tls.hpp>
#include <websocketpp/server.hpp>

typedef websocketpp::server<websocketpp::config::asio> server;

using websocketpp::lib::placeholders::_1;
using websocketpp::lib::placeholders::_2;
using websocketpp::lib::bind;

// pull out the type of messages sent by our config
typedef server::message_ptr message_ptr;

#endif



#define DBG_OUT(x) std::cerr << #x << " = " << x << std::endl
#define EC_CMD_FOUND      (0)
#define EC_CMD_NOT_FOUND (-1)

using boost::asio::ip::tcp;

using namespace log4cxx;
using namespace log4cxx::helpers;

namespace po = boost::program_options;
typedef boost::shared_ptr<tcp::socket> socket_ptr;






LoggerPtr logger(Logger::getLogger("this file")); // replace with name of this file
bool local_kill_switch = false;

std::string command_parser (std::vector<char> command, unsigned int& ec);
void command_session(socket_ptr sock);
void command_server(short port);

void data_session(socket_ptr sock);
void data_server(short port);


extern void read_binary_file_as_cf32(std::vector<std::complex<float> > &dest_vector, std::string source_file_str);

extern void tx_signal_handler(CDeviceStorage& tx_mdst, std::vector<std::complex<float> > &tx_signal, unsigned int run_time, unsigned int intv_us);
extern void tx_signal_handler(CDeviceStorage& tx_mdst, std::vector<std::complex<float> > &tx_signal, unsigned int run_time, unsigned int intv_us, std::string ch_str);
extern void rx_handler_sync_to_pnseq(CDeviceStorage& rx_mdst, std::vector<std::complex<float> > &pnseq, unsigned int run_time, std::vector<std::complex<float> > &outsig);
extern void rx_background_thread_pri255(CDeviceStorage& rx_mdst, std::vector<float>& signal_mag);





int nCmdThreadCnt = 0;
int nDataThreadCnt = 0;

void signal_interrupt_control_c(int s){
  printf("Caught signal %d\n",s);
  local_kill_switch = true;
}


CRadio usrp1;

CDeviceStorage g_rx_mdst;
CDeviceStorage g_tx_mdst;
std::vector<float> g_rx_sig_mag; // instant
unsigned int g_fft_size;

#ifdef _USING_WEBSOCK_

float f32_rand(float a, float b) {
  float random = ((float) rand()) / (float) RAND_MAX;
  float diff = b - a;
  float r = random * diff;
  return a + r;
}

// Define a callback to handle incoming messages
void on_message(server* s, websocketpp::connection_hdl hdl, message_ptr msg) {
  //std::cout << "on_message called with hdl: " << hdl.lock().get()
  //            << " and message: " << msg->get_payload()
  //            << std::endl;
  DBG_OUT( msg->get_payload() );

  // check for a special command to instruct the server to stop listening so
  // it can be cleanly exited.
  if (msg->get_payload() == "stop-listening") {
    s->stop_listening();
    return;
  }

  std::string one_api_string ( msg->get_payload() ) ;
  std::vector<char> command(one_api_string.begin(), one_api_string.end() );
  unsigned int ec = EC_CMD_FOUND;
  command.push_back('\0');
  command_parser( command, ec);

  // quick soln
  // if command was already found above then return
  if (ec == EC_CMD_FOUND) return;


  // command was not found above so look up below...
  // need to add command look up
  try {
    // kludge
    if (g_rx_sig_mag.size() == 0) return;

    //s->send(hdl, msg->get_payload(), msg->get_opcode());
    std::vector<unsigned int> ui32_val;

    ui32_val.resize(2 + g_fft_size);
    ui32_val.at(0) = 0x1;
    //ui32_val.at(1) = 0x0;
    for (unsigned int i = 2; i < (2 + g_fft_size); i++)
      *((float*)&ui32_val.at(i)) = g_rx_sig_mag.at(i-2); //f32_rand(-0.5, 0.5);

    unsigned int plyd_bytes = sizeof(ui32_val.at(0)) * (ui32_val.size() - 2) ; // 2 is for header index: {0,1}
    ui32_val.at(1) = (float)( plyd_bytes ); 
      
    std::string str_msg;
    str_msg.resize(sizeof(ui32_val.at(0)) * ui32_val.size());

    DBG_OUT( str_msg.length() );
    memcpy((void*)str_msg.data(), (void*)ui32_val.data(), str_msg.length() );

    s->send(hdl, str_msg, websocketpp::frame::opcode::binary);
    //s->send(hdl, std::string("hello world"), websocketpp::frame::opcode::text );
  } catch (websocketpp::exception const & e) {
    std::cout << "Echo failed because: "
	      << "(" << e.what() << ")" << std::endl;
  }
}



void websock_server(short port)
{
    // Websocket configuration
    server echo_server; // creates server endpoint
    LOG4CXX_INFO(logger, "Starting websock server at " << port);


    try {
        // Set logging settings
        echo_server.set_access_channels(websocketpp::log::alevel::all);
        echo_server.clear_access_channels(websocketpp::log::alevel::frame_payload);

        // Initialize Asio
        echo_server.init_asio();

	echo_server.set_reuse_addr (true);

        // Register our message handler
        echo_server.set_message_handler(bind(&on_message,&echo_server,::_1,::_2));

        // Listen on port 9002
        echo_server.listen( port );

        // Start the server accept loop
        echo_server.start_accept();

        // Start the ASIO io_service run loop
        echo_server.run();
    } catch (websocketpp::exception const & e) {
        std::cout << e.what() << std::endl;
    } catch (...) {
        std::cout << "other exception" << std::endl;
    }

}
#endif



int main(int argc, char *argv[]){

  short command_server_port;
  short websock_server_port;
  unsigned int intv_ms;
  //std::string api_string;
  std::vector<std::string> api_strings;

    //setup the program options
    po::options_description desc("Allowed options");
    desc.add_options()
      ("help", "help message")
      //("time", po::value<unsigned int>(&run_time)->default_value(10), "run time in seconds")
      ("intv", po::value<unsigned int>(&intv_ms)->default_value(1000),"interval (msec) between transmits")
      ("cmd-port", po::value<short>(&command_server_port)->default_value(5180),"command server port")
      ("ws-port", po::value<short>(&websock_server_port)->default_value(9002),"http command / websock server port")
      ("cmd", po::value<std::vector<std::string> > (&api_strings),"calls a single user defined API")
      ;

    po::variables_map vm;
    po::store(po::parse_command_line(argc, argv, desc), vm);
    po::notify(vm);

    //print the help message
    if (vm.count("help")){
        std::cout << boost::format("This is a test Command Line Player\n %s") % desc << std::endl;
	return 0;
    }

    // Configure the logger
    PropertyConfigurator::configure("./logconf.prop");

    // Ctrl-C for breaking
    signal (SIGINT, signal_interrupt_control_c);

    if (vm.count("cmd")){
      for (unsigned int i = 0; i < api_strings.size(); i++) {
	std::string one_api_string = api_strings.at(i);
	std::vector<char> command(one_api_string.begin(), one_api_string.end() );
	unsigned int ec = EC_CMD_FOUND;
	command.push_back('\0');
	command_parser( command, ec);
      }
      //return 0;
    }

    // launch thread for remote control
    //unsigned int data_server_port = command_server_port + 1;
    boost::thread cmds(command_server, command_server_port);
    //boost::thread data(data_server,    data_server_port);



#ifdef _USING_WEBSOCK_
    boost::thread cmds_http(websock_server, websock_server_port);
#endif


    // Back ground loop
    boost::system_time next_console_refresh = boost::get_system_time();
    while(!local_kill_switch) {
      if (boost::get_system_time() > next_console_refresh) {
	//count_processed_buffers = 0;
	next_console_refresh = boost::get_system_time() + boost::posix_time::microseconds(long(1000e3));
	std::cerr << ".";
      }
    } //  while(!local_kill_switch)

    boost::this_thread::sleep(boost::posix_time::milliseconds(100));
    
    //finished
    std::cerr << std::endl << "Done!" << std::endl << std::endl;
    return EXIT_SUCCESS;
}









boost::thread *p_thread_fft = 0;

#define MAKE_3_STRING_RSP(Arg1,Arg2,Arg3)			\
  Arg1								\
    + " " + boost::lexical_cast<std::string>( Arg2 )		\
    + " " + boost::lexical_cast<std::string>( Arg3 ) + "\n"

std::string command_parser (std::vector<char> command, unsigned int& ec)
{
  std::string temp( &command.front() );
  LOG4CXX_INFO(logger, "Rqst: " + temp);
  std::vector<std::string> sToken;
  boost::split(sToken, temp, boost::is_any_of(" ") );
  
#if 0
  for (unsigned int i = 0; i < sToken.size(); i++)
    std::cerr << sToken.at(i) << " ";
  std::cerr << std::endl;
#endif

  ec = EC_CMD_FOUND;
  temp = "Rsp: ";
  // token size 3 are set functions, size 2 are get functions
  if ( (sToken.size() == 2) && (sToken.at(0) == "make") ) {
    std::string ip_str = sToken.at(1);

    temp += &command.front();
    
    //std::string arg_str = ip_str;
    //usrp1.set_sync_pps();
    
    if (usrp1.make_device_handle( ip_str ) != 0)
      temp += " Device handle already exists --> make ignored!";

  }
  else if ( (sToken.size() == 1) && (sToken.at(0) == "status")  ) {
    temp += &command.front();
    temp += " " + usrp1.mdst_state();

    //temp += "\n";
    //temp += boost::lexical_cast<std::string>(g_rx_sig_mag.size()) + " ";
    //temp += boost::lexical_cast<std::string>(g_rx_sig_mag.at(0)) + " ";
    //temp += boost::lexical_cast<std::string>(g_rx_sig_mag.at(1));

  }
  else if ( (sToken.size() == 1) && (sToken.at(0) == "rxpause")  ) {
    temp += &command.front();
    temp += " pause on";
    usrp1.pause_rx();

  }
  else if ( (sToken.size() == 1) && (sToken.at(0) == "rxcont")  ) {
    temp += &command.front();
    temp += " pause off";
    usrp1.cont_rx();

  }
  else if ( (sToken.size() == 2) && (sToken.at(0) == "rxstream")  && (sToken.at(1) == "start") ) {

    temp += &command.front();
    if (usrp1.run_rx_async() != 0)
      temp = "rxstream already in progress --> command ignored!";

                                 // move line below to rxsig
    g_rx_mdst = usrp1.rx_mdst(); // the receiver's storage structure used by USRP
  }
  else if ( (sToken.size() == 3) && (sToken.at(0) == "rxstream")  && (sToken.at(1) == "start_fft") ) {
    g_fft_size = boost::lexical_cast<unsigned int>( sToken.at(2) ); // usrp1.spb() *32
      // !! check fft_size is integer of spb() and a power of 2
    assert( (g_fft_size % usrp1.spb() ) == 0 );

    temp += &command.front();
    if (usrp1.run_rx_async() != 0) {
      temp += "rxstream already in progress --> command ignored!\n";
    }

    if ( p_thread_fft )
    {
      temp += "fft thread already running - command ignored\n";
    }
    else
    {
      temp += "starting fft thread\n";
      // move line below to rxsig
      g_rx_mdst = usrp1.rx_mdst(); // the receiver's storage structure used by USRP
      g_rx_sig_mag.resize( g_fft_size ); // *4 for 1024      *32 for 8192
      p_thread_fft = new boost::thread(rx_background_thread_pri255, g_rx_mdst, boost::ref(g_rx_sig_mag));
    }
  }
  else if ( (sToken.size() == 2) && (sToken.at(0) == "rxstream")  && (sToken.at(1) == "stop_fft") ) {

    temp += &command.front();
    extern bool rx_background_thread_pri255_kill_switch;
    rx_background_thread_pri255_kill_switch = true;
    free(p_thread_fft);
    p_thread_fft = 0;
  }
  else if ( (sToken.size() == 2) && (sToken.at(0) == "fftch") ) {
    temp += &command.front();
    extern unsigned int channel_select;
    channel_select = boost::lexical_cast<unsigned int>( sToken.at(1));
  }
  else if ( (sToken.size() == 2) && (sToken.at(0) == "rxread") ) {
    unsigned int nSamplesToRead = boost::lexical_cast<unsigned int>( sToken.at(1) );
    temp.clear();
    temp.resize( nSamplesToRead * 8 ); // each sample is 8 bytes

    //while( usrp1.rx_mdst().head() != (usrp1.rx_mdst().nbuffptrs()-10) ) {}
    usrp1.pause_rx();
    
#if 0
    std::vector<std::complex<float> > cf(nSamplesToRead);
    cf.at(0) = std::complex<float>(1.0001,2.0002);
    cf.at(1) = std::complex<float>(3.0003,4.0004);
    cf.at(2) = std::complex<float>(5.0005,6.0006);
#endif

    memcpy((void*)temp.data(), (void*)g_rx_mdst.buffer_ptr_ch_idx_( 0, 0 /*g_rx_mdst.tail( )*/ ), temp.size() );
    //memcpy((void*)temp.data(), (void*)cf.data(), temp.size() );

    usrp1.cont_rx();

    //temp += &command.front();

  }

  else if ( (sToken.size() == 2) && (sToken.at(0) == "rxsig")  ) {
    // Format: rxsig <signal file>
    std::string source_file_str = sToken.at(1);

    temp += &command.front();

    // load signal file into vector
    std::vector<std::complex<float> > ld_signal_cf32;
    read_binary_file_as_cf32(ld_signal_cf32, source_file_str); 
    DBG_OUT(ld_signal_cf32.size());

    std::vector<std::complex<float> > outsig;
    outsig.resize(256*10);
    rx_handler_sync_to_pnseq(g_rx_mdst, ld_signal_cf32, 10 /* runtime seconds */, outsig);

  }
  else if ( (sToken.size() == 5) && (sToken.at(0) == "txsig")  ) {
    // Format: txsig <signal file> <runtime> <intv>
    
    // assign variable for command line tokens
    std::string source_file_str = sToken.at(1);
    unsigned int runtime = boost::lexical_cast<unsigned int>( sToken.at(2) );
    unsigned int intv = boost::lexical_cast<unsigned int>( sToken.at(3) );
    std::string ch_str = sToken.at(4);

    temp += &command.front();

    // load signal file into vector
    std::vector<std::complex<float> > ld_signal_cf32;
    read_binary_file_as_cf32(ld_signal_cf32, source_file_str); 
    DBG_OUT(ld_signal_cf32.size());

    // send signal 
    g_tx_mdst = usrp1.tx_mdst(); // the transmiter's storage structure used by USRP

    boost::thread txsig_thread(boost::bind( tx_signal_handler, g_tx_mdst, ld_signal_cf32, runtime, intv, ch_str) );
  }

  else if ( (sToken.size() == 3) && (sToken.at(0) == "rxfreq")  && (boost::lexical_cast<unsigned int>( sToken.at(1) ) < usrp1.nRxChannels() ) )
  {
    unsigned int ch = boost::lexical_cast<unsigned int>( sToken.at(1) );
    double set_val = boost::lexical_cast<double>( sToken.at(2) );
    double act_val = usrp1.rx_freq(ch, set_val);
    temp += MAKE_3_STRING_RSP(sToken.at(0), ch, act_val);
  }
  else if ( (sToken.size() == 1) && (sToken.at(0) == "rxfreq") ) {
    for (unsigned int i = 0; i < usrp1.nRxChannels(); i++)
    {
      double act_val = usrp1.rx_freq(i);
      temp += MAKE_3_STRING_RSP(sToken.at(0), i, act_val);
    } 
  }
  else if ( (sToken.size() == 3) && (sToken.at(0) == "rxgain")  && (boost::lexical_cast<unsigned int>( sToken.at(1) ) < usrp1.nRxChannels() ) )
  {
    unsigned int ch = boost::lexical_cast<unsigned int>( sToken.at(1) );
    double set_val = boost::lexical_cast<double>( sToken.at(2) );
    double act_val = usrp1.rx_gain(ch, set_val);

    temp += MAKE_3_STRING_RSP(sToken.at(0), ch, act_val);
  }
  else if ( (sToken.size() == 1) && (sToken.at(0) == "rxgain") ) {
    for (unsigned int i = 0; i < usrp1.nRxChannels(); i++)
    {
      double act_val = usrp1.rx_gain(i);
      temp += MAKE_3_STRING_RSP(sToken.at(0), i, act_val);
    } 
  }
  else if ( (sToken.size() == 3) && (sToken.at(0) == "rxrate")  && (boost::lexical_cast<unsigned int>( sToken.at(1) ) < usrp1.nRxChannels() ) )
    {
      unsigned int ch = boost::lexical_cast<unsigned int>( sToken.at(1) );
      double set_val = boost::lexical_cast<double>( sToken.at(2) );
      double act_val = usrp1.rx_rate(ch, set_val);
      temp += MAKE_3_STRING_RSP(sToken.at(0), ch, act_val);
    }
  else if ( (sToken.size() == 1) && (sToken.at(0) == "rxrate")) {
    for (unsigned int i = 0; i < usrp1.nRxChannels(); i++)
    {
      double act_val = usrp1.rx_rate(i);
      temp += MAKE_3_STRING_RSP(sToken.at(0), i, act_val);
    }
  }


  //
  // Receiver chain commands
  //

  // rxconif <chan> <freq> <rate> <gain>
  else if ( (sToken.size() == 5) && (sToken.at(0) == "rxconfig")  && (boost::lexical_cast<unsigned int>( sToken.at(1) ) < usrp1.nRxChannels() ) )
  {

    unsigned int ch = boost::lexical_cast<unsigned int>( sToken.at(1) );
    double set_val_freq = boost::lexical_cast<double>( sToken.at(2) );
    double act_val_freq = usrp1.rx_freq(ch, set_val_freq);

    double set_val_rate = boost::lexical_cast<double>( sToken.at(3) );
    double act_val_rate = usrp1.rx_rate(ch, set_val_rate);

    double set_val_gain = boost::lexical_cast<double>( sToken.at(4) );
    double act_val_gain = usrp1.rx_gain(ch, set_val_gain);

    for (unsigned int ch = 0; ch < usrp1.nRxChannels(); ch++)
    {
      temp += "\n";
      temp += MAKE_3_STRING_RSP(std::string("rxfreq"), ch, usrp1.rx_freq(ch));
      temp += MAKE_3_STRING_RSP(std::string("rxrate"), ch, usrp1.rx_rate(ch));
      temp += MAKE_3_STRING_RSP(std::string("rxgain"), ch, usrp1.rx_gain(ch));
    }

  }
  else if ( (sToken.size() == 1) && (sToken.at(0) == "rxconfig") ) {
    for (unsigned int ch = 0; ch < usrp1.nRxChannels(); ch++)
    {

      temp += "\n";
      temp += MAKE_3_STRING_RSP(std::string("rxfreq") , ch, usrp1.rx_freq(ch));
      temp += MAKE_3_STRING_RSP(std::string("rxrate") , ch, usrp1.rx_rate(ch));
      temp += MAKE_3_STRING_RSP(std::string("rxgain") , ch, usrp1.rx_gain(ch));

    } 
  }



  // 
  // Transmit chain commands
  //

  // txconif <chan> <freq> <rate> <gain>
  else if ( (sToken.size() == 5) && (sToken.at(0) == "txconfig")  && (boost::lexical_cast<unsigned int>( sToken.at(1) ) < usrp1.nTxChannels() ) )
  {

    unsigned int ch = boost::lexical_cast<unsigned int>( sToken.at(1) );
    double set_val_freq = boost::lexical_cast<double>( sToken.at(2) );
    double act_val_freq = usrp1.tx_freq(ch, set_val_freq);
    
    double set_val_rate = boost::lexical_cast<double>( sToken.at(3) );
    double act_val_rate = usrp1.tx_rate(ch, set_val_rate);
    
    double set_val_gain = boost::lexical_cast<double>( sToken.at(4) );
    double act_val_gain = usrp1.tx_gain(ch, set_val_gain);
    
    for (unsigned int ch = 0; ch < usrp1.nTxChannels(); ch++)
    {
      temp += "\n";
      temp += MAKE_3_STRING_RSP(std::string("txfreq") , ch, usrp1.tx_freq(ch));
      temp += MAKE_3_STRING_RSP(std::string("txrate") , ch, usrp1.tx_rate(ch));
      temp += MAKE_3_STRING_RSP(std::string("txgain") , ch, usrp1.tx_gain(ch));
    }
  }
  else if ( (sToken.size() == 1) && (sToken.at(0) == "txconfig") ) {
    for (unsigned int ch = 0; ch < usrp1.nTxChannels(); ch++)
    {

      temp += "\n";
      temp += MAKE_3_STRING_RSP(std::string("txfreq") , ch, usrp1.tx_freq(ch));
      temp += MAKE_3_STRING_RSP(std::string("txrate") , ch, usrp1.tx_rate(ch));
      temp += MAKE_3_STRING_RSP(std::string("txgain") , ch, usrp1.tx_gain(ch));
    } 
  }
  else if ( (sToken.size() == 2) && (sToken.at(0) == "txstream") && (sToken.at(1) == "start") ) {

    temp += &command.front();
    if (usrp1.run_tx_async() != 0)
      temp = "txstream already in progress --> command ignored!";

    //temp += &command.front();
    //usrp1.run_tx_async();
    g_tx_mdst = usrp1.tx_mdst();

  }
  else if ( (sToken.size() == 2) && (sToken.at(0) == "txstream") && (sToken.at(1) == "stop") ) {
    temp += &command.front();
    usrp1.kill_tx();
  }


  else {
    temp += "---Unknown command---";
    ec = EC_CMD_NOT_FOUND;
  }

  LOG4CXX_INFO(logger, temp);
  return (temp);
}





void command_session(socket_ptr sock)
{
  boost::system::error_code error;
 
  try
  { 
    while(1)
    {
      std::vector<char> command(256*100);

      sock->read_some(boost::asio::buffer( command ), error); // wait for the client to query
      if(error == boost::asio::error::eof)
      {
        LOG4CXX_INFO(logger, "Connection closed");
        break;
      }
      else if(error)
	throw boost::system::system_error(error);
      //LOG4CXX_INFO(logger, (char*)&command.front());

      std::string resp;
      unsigned int ec;
      resp += command_parser(command, ec);
      //resp += "\r\n";

      boost::asio::write(*sock, boost::asio::buffer(resp.data(),resp.size() ));
	
    }
    --nCmdThreadCnt;
    //LOG4CXX_INFO(logger, "command thread count: " << nCmdThreadCnt << " [host: " << sock->remote_endpoint().address().to_string() << "]" );
  }  
  catch (std::exception& e)
  {
    std::cerr << "Exception in command thread: " << e.what() << "\n";
    LOG4CXX_ERROR(logger, "command thread count: " << --nCmdThreadCnt << " [host: " << sock->remote_endpoint().address().to_string() << "]" );
    return;
  }
}


void command_server(short port)
{
  boost::asio::io_service io_service;

  LOG4CXX_INFO(logger, "Starting command server at " << port);
  tcp::acceptor a(io_service, tcp::endpoint(tcp::v4(), port));
  for (;;)
  {
    socket_ptr sock(new tcp::socket(io_service));
    if (nCmdThreadCnt >= 1) continue;
    a.accept(*sock);
    boost::thread t(boost::bind(command_session, sock));
    ++nCmdThreadCnt;
    //LOG4CXX_INFO(logger, "command thread count: " << nCmdThreadCnt << " [host: " << sock->remote_endpoint().address().to_string() << "]" );
  }
}






void data_session(socket_ptr sock)
{
  boost::system::error_code error;
 
  try
  { 
    while(1)
    {
      std::vector<char> command(256*100);

      //sock->read_some(boost::asio::buffer( command ), error); // wait for the client to query
      if(error == boost::asio::error::eof)
      {
        LOG4CXX_INFO(logger, "Connection closed");
        break;
      }
      else if(error)
	throw boost::system::system_error(error);
      //LOG4CXX_INFO(logger, (char*)&command.front());

      std::vector<unsigned char> uc8_resp;
      //uc8_resp = command_parser(command);

      boost::asio::write(*sock, boost::asio::buffer(uc8_resp.data(),uc8_resp.size() ));

    }
    --nCmdThreadCnt;
    //LOG4CXX_INFO(logger, "command thread count: " << nCmdThreadCnt << " [host: " << sock->remote_endpoint().address().to_string() << "]" );
  }  
  catch (std::exception& e)
  {
    std::cerr << "Exception in command thread: " << e.what() << "\n";
    LOG4CXX_ERROR(logger, "command thread count: " << --nCmdThreadCnt << " [host: " << sock->remote_endpoint().address().to_string() << "]" );
    return;
  }
}


void data_server(short port)
{
  boost::asio::io_service io_service;

  LOG4CXX_INFO(logger, "Starting data server at " << port);
  tcp::acceptor a(io_service, tcp::endpoint(tcp::v4(), port));
  for (;;)
  {
    socket_ptr sock(new tcp::socket(io_service));
    if (nDataThreadCnt >= 1) continue;
    a.accept(*sock);
    boost::thread t(boost::bind(data_session, sock));
    ++nDataThreadCnt;
    //LOG4CXX_INFO(logger, "command thread count: " << nDataThreadCnt << " [host: " << sock->remote_endpoint().address().to_string() << "]" );
  }
}

