#include <iostream>
#include <complex>
#include <vector>
#include <boost/lexical_cast.hpp>
#include "CDeviceStorage.hpp"
#include "C1DCorrelation.hpp"

#include <boost/thread.hpp>

#include <fftw3.h>


#include "log4cxx/logger.h"
#include "log4cxx/propertyconfigurator.h"
#include "log4cxx/helpers/exception.h"



//#define WITH_OML
#ifdef WITH_OML
#include "CWriteOml.h"
#endif

using namespace log4cxx;
using namespace log4cxx::helpers;


extern const std::string currentDateTime();
extern void plot_cf32_buffer(std::string udp_dst_addr, std::string port, void *ptr_cf32, unsigned int n_cf32_values);
extern void read_binary_file_as_cf32(std::vector<std::complex<float> > &dest_vector, std::string source_file_str);
extern void write_cf32_to_binary_file(std::string dest_file_str, void *ptr_cf32, unsigned int n_cf32_values);

static bool local_kill_switch = false;
bool rx_background_thread_pri255_kill_switch = false;



unsigned int channel_select = 0;
void rx_background_thread_pri255(CDeviceStorage& rx_mdst, std::vector<float>& signal_mag)
{
#ifdef WITH_OML
  std::string server_port, db_filename;
  server_port = "oml:3003";
  db_filename = "COSMOS_OMLBLOB";
  double d64;
  CWriteOml oml(db_filename, server_port );
  //oml.init(db_filename, server_port );
  boost::this_thread::sleep(boost::posix_time::milliseconds(1000));

  std::vector< std::pair<std::string, OmlValueT> > _omlKeys; 

  _omlKeys.push_back( std::make_pair("CenterFreqMHz", OML_DOUBLE_VALUE));
  _omlKeys.push_back( std::make_pair("SampleRateMHz", OML_DOUBLE_VALUE));
  _omlKeys.push_back( std::make_pair("GainDB",        OML_DOUBLE_VALUE));

  _omlKeys.push_back( std::make_pair("TheBlob",     OML_BLOB_VALUE));
  
  oml.start( _omlKeys );

  d64 = (double)2440e6 / 1e6;
  oml.set_key("CenterFreqMHz", (void*)&d64 );

  d64 = (double)12e6 / 1e6;
  oml.set_key("SampleRateMHz", (void*)&d64 );

  d64 = (double)20;
  oml.set_key("GainDB", (void*)&d64 );

#endif


#if 0
  for(unsigned int cnt = 0; cnt< 3;cnt++) {

    // TMP float array
    float f32[16];
    for (unsigned int i = 0; i<16; f32[i] = (float)cnt+i+1,++i);

   
    oml.set_key_blob("TheBlob", (void*)&f32[0],  16 * sizeof(float));
    oml.insert();

    boost::this_thread::sleep(boost::posix_time::milliseconds(1000));

  }
  //oml.stop(); // destructor will be called when going out of scope

  return;
#endif


  
  // Configure the logger
  LoggerPtr logger(Logger::getLogger("rxtx_handlers.cpp")); // replace with name of this file
  PropertyConfigurator::configure("./logsig.prop");

  unsigned int spb = rx_mdst.spb();
  //unsigned int nRxChannels = rx_mdst.nChannels();
  //unsigned int total_num_samps = rx_mdst.nsamps_per_ch();
  unsigned int n_spb_buffs = rx_mdst.nbuffptrs();
  DBG_OUT( spb );
  DBG_OUT( n_spb_buffs );
  DBG_OUT( signal_mag.size() );

  for (unsigned int i = 0; i < signal_mag.size(); i++)
    signal_mag.at(i) = 0.0;

  std::vector<float> signal_mag_instant( signal_mag.size() );
  std::vector<float> signal_mag_average( signal_mag.size() );

  std::vector<std::complex<float> > time_buff( signal_mag.size() );
  std::vector<std::complex<float> > freq_buff( signal_mag.size() );
  unsigned int accum_depth = signal_mag.size() / spb;
  DBG_OUT(accum_depth);
  assert( freq_buff.size() == spb*accum_depth);

  fftwf_plan fft_p = fftwf_plan_dft_1d( signal_mag.size(),
					(fftwf_complex*)&time_buff.front(),
					(fftwf_complex*)&freq_buff.front(),
					FFTW_FORWARD,
					FFTW_ESTIMATE);

  boost::system_time next_console_refresh = boost::get_system_time() + boost::posix_time::microseconds(long(1.0e6));

  rx_background_thread_pri255_kill_switch = false;
  unsigned int avg_count = 0;
  unsigned int AVG_WINDOW = 64;
  while ( !rx_background_thread_pri255_kill_switch ) {

    if (boost::get_system_time() < next_console_refresh) continue;
    next_console_refresh = boost::get_system_time() + boost::posix_time::microseconds(long(1));


    if (rx_mdst.depth() < accum_depth ) continue;

    memcpy((void*)time_buff.data(),
	   (void*)rx_mdst.buffer_ptr_ch_idx_( channel_select, rx_mdst.tail( ) ),
	   signal_mag.size() * 1 * sizeof(std::complex<float>) );
    
    fftwf_execute(fft_p); // FWD FFT is NOT normalized so divide by N

    // magnitude - ignore frequencies around DC component
    for(unsigned int i = 0; i < signal_mag.size(); i++)
      signal_mag_instant.at(i) = abs(freq_buff.at(i));

    
    for(unsigned int i = 0; i < signal_mag.size(); i++)
      signal_mag_average.at(i) += signal_mag_instant.at(i);

    avg_count++;
    if (avg_count % AVG_WINDOW)
    {
      for(unsigned int i = 0; i < signal_mag.size(); i++)
	signal_mag_average.at(i) /= AVG_WINDOW;
    
      signal_mag = signal_mag_average;

      for (unsigned int i = 0; i < signal_mag.size(); i++)
	signal_mag_average.at(i) = 0.0;

    }

    rx_mdst.tail_set( rx_mdst.tail( accum_depth ) );
    //rx_mdst.head_set( 0 );
    //rx_mdst.tail_set( 0 );

#ifdef WITH_OML
    oml.set_key_blob("TheBlob", (void*)&signal_mag.front(), (unsigned int)signal_mag.size() * sizeof(float) );
    oml.insert();
#endif


  } // while



#ifdef WITH_OML
  //oml.stop(); // destructor will be called when going out of scope
#endif
  
  LOG4CXX_INFO(logger, std::string(__FUNCTION__) + " - exitting...");
  
} // rx_background_thread_pri255






void rx_handler_sync_to_pnseq(CDeviceStorage& rx_mdst, std::vector<std::complex<float> > &pnseq, unsigned int run_time, std::vector<std::complex<float> > &outsig) {
  // Configure the logger
  LoggerPtr logger(Logger::getLogger("rxtx_handlers.cpp")); // replace with name of this file
  PropertyConfigurator::configure("./logsig.prop");

  //boost::this_thread::sleep(boost::posix_time::milliseconds(100));

  unsigned int spb = rx_mdst.spb() /* / 4*/;
  unsigned int nRxChannels = rx_mdst.nChannels();
  unsigned int total_num_samps = rx_mdst.nsamps_per_ch();
  unsigned int n_spb_buffs = rx_mdst.nbuffptrs();

  std::vector<float> mag_buff( spb );
  std::vector<std::complex<float> > fft_buff( spb );
  assert( fft_buff.size() == spb);

  boost::system_time done_time = boost::get_system_time() + boost::posix_time::microseconds(long( run_time*1.0e6) );

  // sync stuff

  C1DCorrelation corr1d_pn;

  // Initialize correlator with time domain sync signal
  unsigned int sync_length = pnseq.size();
  unsigned int signal_length = spb*10; //spb*4;
  corr1d_pn.init( pnseq,       // sync vector samples
	          sync_length,        // sync vector length
	          signal_length);     // signal length

  //rx_mdst.pause(false);

  unsigned int trigger_ch = 0; // this is the channel to trigger on
  local_kill_switch = false;
  unsigned int accum_depth = 10;
  std::string g_signal_out_file = "tmp_" + currentDateTime() + ".out";

  while ((!local_kill_switch) && (boost::get_system_time() < done_time)) {

    // wait to fill entire buffer for channel then continue
    if (rx_mdst.depth() < accum_depth ) continue;

    LOG4CXX_INFO(logger, "ch | depth | tail | head : " << trigger_ch << " " << rx_mdst.depth() << " " << rx_mdst.tail() << " " << rx_mdst.head() );

    if (rx_mdst.depth() == 10) {
      unsigned int corr_idx = corr1d_pn.run( (fftwf_complex*) rx_mdst.buffer_ptr_ch_idx_( 0, rx_mdst.tail() ) );
      DBG_OUT( corr1d_pn.corr_idx() );
      DBG_OUT( corr1d_pn.corr_mag() );
      if (corr1d_pn.corr_mag() > 100.0) {
	//plot_cf32_buffer("10.10.0.10", "1337",  (std::complex<float> *)corr1d_pn.corr_magnitudes().data(), corr1d_pn.corr_magnitudes().size() );
	plot_cf32_buffer("10.10.0.10", "1337",  (std::complex<float> *)rx_mdst.buffer_ptr_ch_idx_( 0, rx_mdst.tail( ) ), spb * 10);

	//write_cf32_to_binary_file(g_signal_out_file,(std::complex<float> *)rx_mdst.buffer_ptr_ch_idx_( 0, rx_mdst.tail( ) ) /* + corr_idx_sp.at(0)*/, spb * 10);

	memcpy((void*)outsig.data(), (void*)rx_mdst.buffer_ptr_ch_idx_( 0, rx_mdst.tail( ) ), spb * 10 * sizeof(std::complex<float>) );
	
	local_kill_switch = true;
      }
    }

    rx_mdst.tail_set( rx_mdst.tail( accum_depth ) );
    //rx_mdst.head_set( 0 );
    //rx_mdst.tail_set( 0 );

  } // while

  std::cerr << "rx signal handler - exitting..." << std::endl;

} // rx_handler_sync_to_pnseq



void tx_signal_handler(CDeviceStorage& tx_mdst, std::vector<std::complex<float> > &tx_signal, unsigned int run_time, unsigned int intv_us, std::string ch_str)
{
  unsigned int spb = tx_mdst.spb();
  unsigned int nTxChannels = tx_mdst.nChannels();
  unsigned int total_num_samps = tx_mdst.nsamps_per_ch();
  unsigned int n_spb_buffs = tx_mdst.nbuffptrs();
  //DBG_OUT(spb); DBG_OUT(nTxChannels); DBG_OUT(total_num_samps); DBG_OUT (n_spb_buffs);

  boost::system_time done_time = boost::get_system_time() + boost::posix_time::microseconds(long( run_time*1.0e6) );

  if (ch_str == "ALL") {
    // copy signal to tx_mdst buffers for sending
    assert(tx_signal.size() <= total_num_samps );
    for (unsigned int ch = 0; ch < nTxChannels; ch++) {
      memcpy((void*)tx_mdst.buffer_ptr_ch_idx_(ch, 0 /*idx*/),
	     (void*)tx_signal.data(), tx_signal.size() * sizeof(std::complex<float>) );
    }
  }
  else {
    unsigned int ch = boost::lexical_cast<unsigned int>( ch_str );
    // copy signal to tx_mdst buffers for sending
    assert(tx_signal.size() <= total_num_samps );
    for (unsigned int ch = 0; ch < nTxChannels; ch++) {
      memset((void*)tx_mdst.buffer_ptr_ch_idx_(ch, 0 /*idx*/),
	     0, tx_signal.size() * sizeof(std::complex<float>) );
    }
    memcpy((void*)tx_mdst.buffer_ptr_ch_idx_(ch, 0 /*idx*/),
	   (void*)tx_signal.data(), tx_signal.size() * sizeof(std::complex<float>) );

  }

#if 0 // this method copies in chunks of 256 std::complex<floats>
  for (unsigned int ch = 0; ch < nTxChannels; ch++) {
    // replicate signal to all (N_SPB_BUFFS) buffers per chan
    for (unsigned int idx = 0; idx < /*1*/n_spb_buffs; idx++) {
      memcpy((void*)tx_mdst.buffer_ptr_ch_idx_(ch, idx),
	     (void*)tx_signal.data(), spb * sizeof(std::complex<float>) );
    }
  }
#endif

  while ((!local_kill_switch) && (boost::get_system_time() < done_time)) {
    if (tx_mdst.head() != 0) continue; // flow control here

    // if signal is constant then generate once above the while conditional.
    // if sign is NOT constant then update signal here and copy to tx_mdst buffers.
    // To be tried out :P

    // Instead of sleep use time check method.
    // halt between  signal burst. 
    if (intv_us != 0)
      boost::this_thread::sleep(boost::posix_time::microseconds(intv_us));

    tx_mdst.head_set( n_spb_buffs ); // kick off sending
    //std::cerr << "^";
  }

  std::cerr << "tx signal handler - exitting..." << std::endl;

} // tx_signal_handler


